/*
 * heat2d-e.c -- solve the heat equation in a two dimensional square region
 * by explicit finite difference method. (version 2)
 *
 *      To compile this program on WS's in 6701,
 *              gcc -o heat2d-e-en heat2d-e-en.c -lmatrix -lglscd -lX11 -lm
 *      or try
 *              ccmg heat2d-e.c
 */

#include <stdio.h>
#include <math.h>

/* to use matrix, new_matrix() */
#include <matrix.h>

/* */
#include "call_gnuplot.h"

double f(double, double);
double maxnorm(int, int, matrix);

int
main()
{
    int Nr, Nt, i, j, k, skip, Kmax;
    double hr, ht, ri, theta_j;
    double lambda_r, lambda_t, lambda;
    double tau, Tmax, t, dt, s;
    matrix u, newu;
    double f(double, double);
    char label[100];

    printf("Nr, Nt: ");
    scanf("%d %d", &Nr, &Nt);
    if ((u = new_matrix(Nr + 1, Nt + 1)) == NULL) {
	fprintf(stderr, " u ݤǤޤǤ");
	exit(1);
    }
    if ((newu = new_matrix(Nr + 1, Nt + 1)) == NULL) {
	fprintf(stderr, " newu ݤǤޤǤ");
	exit(1);
    }
    printf("Tmax: ");
    scanf("%lf", &Tmax);

    hr = 1.0 / Nr;
    ht = 2 * M_PI / Nt;

    printf("(%g): ",
           0.5 * (hr * hr * ht * ht) / (1 + ht * ht));
    scanf("%lf", &tau);

    lambda_r = tau / (hr * hr);
    lambda_t = tau / (ht * ht);
    lambda = lambda_r + lambda_t;

    printf("=%g ˤʤޤ\n", lambda);

    printf("t: ");
    scanf("%lf", &dt);
    if (dt < tau) {
	dt = tau;
	printf("dt=%g\n", dt);
    }
    skip = dt / tau + 0.5;

    open_gnuplot();

    for (i = 0; i <= Nr; i++) {
	ri = i * hr;
	for (j = 0; j <= Nt; j++) {
	    theta_j = j * ht;
	    u[i][j] = f(ri * cos(theta_j), ri * sin(theta_j));
	}
    }

    Kmax = Tmax / tau + 0.5;
    for (k = 1; k <= Kmax; k++) {
	for (i = 1; i < Nr; i++) {
	    ri = i * hr;
	    for (j = 0; j < Nt; j++) {
		int jm1 = j - 1;
		if (jm1 == -1)
		    jm1 = Nt - 1;
		newu[i][j] = (1 - 2 * lambda_r - 2 * lambda_t / (ri * ri)) * u[i][j]
		    + lambda_r * (u[i + 1][j] + u[i - 1][j])
		    + lambda_r * hr / (2 * ri) * (u[i + 1][j] - u[i - 1][j])
		    + lambda_t / (ri * ri) * (u[i][j + 1] + u[i][jm1]);
	    }
	}
	for (i = 1; i < Nr; i++)
	    newu[i][Nt] = newu[i][0];
	/* Dirichlet  */
	for (j = 0; j <= Nt; j++)
	    newu[Nr][j] = 0.0;

	s = 0;
	for (j = 0; j < Nt; j++)
	    s += u[1][j];

	newu[0][0] = lambda_r * (4 * (s / Nt - u[0][0])) + u[0][0];
	for (j = 1; j <= Nt; j++)
	    newu[0][j] = newu[0][0];

	for (i = 0; i <= Nr; i++)
	    for (j = 0; j <= Nt; j++)
		u[i][j] = newu[i][j];

	printf("k=%d, norm=%g\n", k, maxnorm(Nr + 1, Nt + 1, u));
	if (k % skip == 0) {
	    t = tau * k;
	    sprintf(label, "t=%g", t);
	    disk(Nr, Nt, u, label);
	}
    }

    close_gnuplot();

    return 0;
}

double f(double x, double y)
{
    double s = 1 - x * x - y * y;
    if (s < 0.0) {
	printf("f(): s= %g", s);
	s = 0.0;
    }
    return sqrt(s);
}

double maxnorm(int m, int n, matrix u)
{
    int i, j, i0, j0;
    double tmpmax, absu;
    i0 = 0;
    j0 = 0;
    tmpmax = fabs(u[0][0]);
    for (i = 0; i < m; i++)
	for (j = 0; j < n; j++)
	    if ((absu = u[i][j]) > tmpmax) {
		tmpmax = absu;
		i0 = i;
		j0 = j;
	    }
    printf("(i,j)=(%d,%d)", i0, j0);
    return tmpmax;
}
